def of_type(args):
    return type(args[0]) == args[1]


def list_(args):
    return of_type([args[0], list])


def bool_(args):
    return of_type([args[0], bool])


def str_(args):
    return of_type([args[0], str])


def int_(args):
    return of_type([args[0], int])


def tuple_(args):
    return of_type([args[0], tuple])


def float_(args):
    return of_type([args[0], float])


def dict_(args):
    return of_type([args[0], dict])


def set_(args):
    return of_type([args[0], set])



import types


class Rule:
    def __init__(self, name: str, fn: types.FunctionType, args: list, invert: bool, every: bool):
        self.name = name
        self.fn = fn
        self.args = args
        self.invert = invert
        self.every = every

    def test(self, value):
        if not self.every:
            return self._test_single_value(value)
        else:
            for elem in value:
                if not self._test_single_value(elem):
                    return False
            return True

    def _test_single_value(self, value):
        test_args = self.args.copy()
        test_args.insert(0, value)
        if not self.invert:
            return self.fn(test_args)
        else:
            return not self.fn(test_args)



from .Rule import Rule
from .rules import *


class Validator:
    def __init__(self, chain: list, invert: bool, every: bool):
        self.chain = chain
        self.invert = invert
        self._every = every

    def test(self, *values):
        for value in values:
            for rule in self.chain:
                if not rule.test(value):
                    return False
        return True

    def not_(self):
        return Validator(self.chain, True, False)
    
    def every(self):
        return Validator(self.chain, False, True)
        

    def greater_than(self, a):
        self.chain.append(Rule("be greater than %s" % a, greater_than, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def greater_or_equal_than(self, a):
        self.chain.append(Rule("be greater or equal than %s" % a, greater_or_equal_than, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def between(self, a, b):
        self.chain.append(Rule("be between %s and %s (inclusive)" % (a, b), between, [a, b], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def str_(self):
        self.chain.append(Rule("be a str", str_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def int_(self):
        self.chain.append(Rule("be an int", int_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def even(self):
        self.chain.append(Rule("be even", even, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def odd(self):
        self.chain.append(Rule("be odd", odd, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def equal(self, a):
        self.chain.append(Rule("be equal to %s" % a, equal, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def min_length(self, a):
        self.chain.append(Rule("be of min length %s" % a, min_length, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def max_length(self, a):
        self.chain.append(Rule("be of max length %s" % a, max_length, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def length(self, a):
        self.chain.append(Rule("be of length %s" % a, length, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def none(self):
        self.chain.append(Rule("be none", none, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def first(self, a):
        self.chain.append(Rule("contain first %s" % a, first, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def last(self, a):
        self.chain.append(Rule("contain last %s" % a, last, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def list_(self):
        self.chain.append(Rule("be a list", list_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def bool_(self):
        self.chain.append(Rule("be a bool", bool_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def less_than(self, a):
        self.chain.append(Rule("be less than %s" % a, less_than, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def less_or_equal_than(self, a):
        self.chain.append(Rule("be less or equal than %s" % a, less_or_equal_than, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def includes(self, a):
        self.chain.append(Rule("include %s" % a, includes, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def positive(self):
        self.chain.append(Rule("be positive", positive, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def negative(self):
        self.chain.append(Rule("be negative", negative, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def empty(self):
        self.chain.append(Rule("be empty", empty, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def divisible_by(self, a):
        self.chain.append(Rule("be divisible by %s" % a, divisible_by, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def float_(self):
        self.chain.append(Rule("be a float", float_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def dict_(self):
        self.chain.append(Rule("be a dict", dict_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def set_(self):
        self.chain.append(Rule("be a set", set_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def tuple_(self):
        self.chain.append(Rule("be a tuple", tuple_, [], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def of_type(self, a):
        self.chain.append(Rule("of type %s" % a, of_type, [a], self.invert, self._every))
        return Validator(self.chain, False, self._every)

    def validate(self, value, value_name: str = None):
        error_message = "%s must:\n" % value_name if value_name is not None else "value must:\n"
        should_raise_value_error = False
        for rule in self.chain:
            if not rule.test(value):
                error_message += "\t- %s" % rule.name
                should_raise_value_error = True
        if should_raise_value_error:
            raise ValueError(error_message)



def greater_than(args):
    return args[0] > args[1]


def greater_or_equal_than(args):
    return args[0] >= args[1]


def between(args):
    return args[1] <= args[0] <= args[2]


def even(args):
    return divisible_by([args[0], 2])


def odd(args):
    return args[0] % 2 != 0


def equal(args):
    return args[0] == args[1]


def min_length(args):
    return len(args[0]) >= args[1]


def max_length(args):
    return len(args[0]) <= args[1]


def length(args):
    return len(args[0]) == args[1]


def none(args):
    return args[0] is None


def first(args):
    return args[0][0] == args[1]


def last(args):
    last_idx = len(args[0]) - 1
    return args[0][last_idx] == args[1]


def less_than(args):
    return args[0] < args[1]


def less_or_equal_than(args):
    return args[0] <= args[1]


def includes(args):
    return args[1] in args[0]


def positive(args):
    return greater_or_equal_than([args[0], 0])


def negative(args):
    return less_than([args[0], 0])


def empty(args):
    return length([args[0], 0])


def divisible_by(args):
    return args[0] % args[1] == 0



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py-v8n",
    version="0.0.5",
    author="Nicolas A. Schejtman",
    author_email="nschejtman93@gmail.com",
    description="Python fluent validation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nschejtman/py-v8n",
    packages=['py_v8n', 'py_v8n.rules'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
